-- apcontrol

-- Copyright (c) 2007 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.

-- control interface for various media players
-- start up /apcontrol
-- configuration is done via the config dialog
-- currently suppported players are: xmms, vlc, mpd

local DEFAULTPLAYER = "xmms"
local DEFAULTADDRESS = "localhost"
local DEFAULTPORT = 1586
local DEFAULTTIMEOUT = 2000
local DEFAULTDUCKING = 0

local function compare_tables(table1, table2)
	if #table1 ~= #table2 then
		return false
	end
	
	for i,v in ipairs(table1) do
		if table2[i] ~= v then
			return false
		end
	end
	return true
end

local function time_to_str(time)
	time = time or 0
	return string.format("%02i:%02i", time/60, time % 60)
end


local apcontrol = {}

local function LoadPlayer(name)
	local success, player
	success, player = pcall(function() return dofile('plugins/apcontrol/players/'..tostring(name)..".lua") end)
	if success then
		return player
	end
end

-- set/change the current player
function apcontrol:SetPlayer(name)
	-- maybe I should go back to loading everything on startup..
	local player = LoadPlayer(name)
	if player then
		self.player = player
		self.playername = name

		-- reinit callbacks
		self:CBInit()
		gkini.WriteString("apcontrol", "player", name)
	else
		print("Error! invalid player: "..tostring(name))
	end
end

function apcontrol:SetPlaylist(pl)
	local old_pl = {}
	
	local i = 1
	while self.playlist[i] do
		table.insert(old_pl, self.playlist[i])
		i = i + 1
	end

	if not compare_tables(old_pl, pl) then
		for i,v in ipairs(pl) do
			self.playlist[i] = v
		end
		self.playlist[#pl+1] = nil
	end
end

-- configuration
function apcontrol:SetConfig()
	gkini.WriteInt("apcontrol", "port", self.port)
	gkini.WriteString("apcontrol", "address", self.address)
	gkini.WriteInt("apcontrol", "timeout", self.timeout)
	if self.ducking then
		gkini.WriteInt("apcontrol", "relayducking", 1)
	else
		gkini.WriteInt("apcontrol", "relayducking", 0)
	end
end

function apcontrol:GetConfig()
	self.playername = gkini.ReadString("apcontrol", "player", DEFAULTPLAYER)
	self.port = gkini.ReadInt("apcontrol", "port", DEFAULTPORT)
	self.address = gkini.ReadString("apcontrol", "address", DEFAULTADDRESS)
	self.timeout = gkini.ReadInt("apcontrol", "timeout", DEFAULTTIMEOUT)
	self.ducking = (gkini.ReadInt("apcontrol", "relayducking", DEFAULTDUCKING) == 1)
end

-- initializes callbacks
function apcontrol:CBInit()
	if self.player then
		function self.player.Title_cb(name) 
			self.track.title = name
		end

		function self.player.Length_cb(length) 
			self.length.title = time_to_str(length)
		end

		function self.player.Time_cb(time)
			self.time.title = time_to_str(time)
		end
	
		function self.player.Playlist_cb(pl)
			self:SetPlaylist(pl)
		end
		
		function self.player.StatusMsg_cb(msg)
			self.status.title = "Status: "..msg 
		end
	end
end

-- create and init config dialog
function apcontrol:CFGInit()
	playerentry = iup.text{value=self.playername, size="THIRD"}
	addrentry = iup.text{value=self.address, size="THIRD"}
	portentry = iup.text{value=self.port, size="THIRD"}
	timeoutentry = iup.text{value=self.timeout, size="THIRD"}

	local close = iup.stationbutton{title="Close", expand="HORIZONTAL"}

	local relayducking = iup.stationtoggle{title="Relay voicechat sound ducking"}
	relayducking.value = self.ducking and "ON" or "OFF"

	local rootbox = iup.vbox{iup.label{title="Config", expand="HORIZONTAL", alignment="ACENTER"},
				 iup.hbox{iup.label{title="Player", expand="HORIZONTAL"}, playerentry},
				 iup.hbox{iup.label{title="Address", expand="HORIZONTAL"}, addrentry},
				 iup.hbox{iup.label{title="Port", expand="HORIZONTAL"}, portentry},
				 iup.hbox{iup.label{title="Timeout", expand="HORIZONTAL"}, timeoutentry},
				 relayducking,
				 iup.fill{expand="YES"},
				 close,
				 gap=2}
	
	self.cfgdialog = iup.dialog{iup.hbox{iup.fill{},
			                           iup.vbox{iup.fill{},
							    iup.stationhighopacityframe{iup.stationhighopacityframebg{rootbox}, expand="NO"},
							    iup.fill{}},
						   iup.fill{}},
					  RESIZE="NO",
					  FULLSCREEN="YES",
					  MENUBOX="NO",
					  TOPMOST="YES",
					  BORDER="NO",
					  BGCOLOR="0 0 0 50 *",
					  DEFAULTESC=close,
					  DEFAULTENTER=close}
				  
	function close.action()
		self.playername = playerentry.value
		self.port = tonumber(portentry.value)
		self.address = addrentry.value
		self.timeout = tonumber(timeoutentry.value)
		self.ducking = (relayducking.value == "ON")

		self:SetConfig()
		self:SetPlayer(self.playername)
		self.cfgdialog:hide()
		self:Show()
	end
end

-- create and init main window
function apcontrol:Init()
	self.track = iup.label{title="--", expand="HORIZONTAL"}
	self.time = iup.label{title="00:00"}
	self.length = iup.label{title="00:00"}
	self.status = iup.label{title="--", expand="HORIZONTAL"}
	self.playlist = iup.stationsubsublist{expand="YES", size="HALF"}

	local config = iup.stationbutton{title="Config", action=function() self:Hide() self.cfgdialog:show() end}
	local close = iup.stationbutton{title="Close", action=function() self:Hide() end, expand="HORIZONTAL"}
	local play = iup.stationbutton{title="Play", action=function() self.player:Play() end, expand="HORIZONTAL"}
	local stop = iup.stationbutton{title="Stop", action=function() self.player:Stop() end, expand="HORIZONTAL"}
	local next = iup.stationbutton{title="Next", action=function() self.player:Next() end, expand="HORIZONTAL"}
	local prev = iup.stationbutton{title="Prev", action=function() self.player:Prev() end, expand="HORIZONTAL"}


	local function volume_cb(ctl, val, idx)
		if idx > 1 then
			self.player:Volume((idx - 2) * 20)
		end
		ctl.value = 1
	end
	local volume = iup.list{dropdown="YES", action=volume_cb}
	
	volume[1] = "Volume"
	volume[2] = "0%"
	volume[3] = "20%"
	volume[4] = "40%"
	volume[5] = "60%"
	volume[6] = "80%"
	volume[7] = "100%"


	-- messy
	local controls1 = iup.hbox{play, stop, prev, next, expand="HORIZONTAL"}
	local controls2 = iup.hbox{config, close, volume, gap=2, ALIGNMENT="ACENTER"}

	local playerbox = iup.vbox{iup.label{title="apcontrol", expand="HORIZONTAL", alignment="ACENTER"},
		                   iup.hbox{iup.vbox{iup.label{title="Track: "},
						     iup.label{title="Time: "}}, 
					    iup.vbox{self.track, 
						     iup.hbox{self.time, iup.label{title=" / "}, self.length}}},
				            iup.stationsubhollowframe{iup.vbox{self.playlist,
							                       controls1,
								               controls2}},
					    self.status,
				   gap=2}

	self.ctldialog = iup.dialog{iup.hbox{iup.fill{},
			                     iup.vbox{iup.fill{},
						      iup.stationhighopacityframe{iup.stationhighopacityframebg{playerbox}, size="HALFxHALF", expand="NO"},
						      iup.fill{}},
					     iup.fill{}},
				    RESIZE="NO",
				    FULLSCREEN="YES",
				    MENUBOX="NO",
				    TOPMOST="YES",
				    BORDER="NO",
				    BGCOLOR="0 0 0 50 *",
				    DEFAULTESC=close,
				    DEFAULTENTER=play}

	function self.playlist.action(_, _, i, on)
		if on == 1 then
			self.player:PlayTrack(i-1)
			local timer = Timer{}
		end
	end

	self:CFGInit()
	self:CBInit()
end


function apcontrol:Show()
	self:SetPlayer(self.playername)
	self.player:Connect(self.address, self.port, self.timeout, true)
	self.ctldialog:show()
end

function apcontrol:Hide()
	self.ctldialog:hide()
	self.player:Disconnect()
end


apcontrol:GetConfig()
apcontrol:Init()


-- command interface

-- vlc doesn't handle multiple connections. don't accept commands when recently locked
local lock = 0
local ACTION_TIMEOUT = 2000

local function check_lock()
	if os.time() - lock <= math.ceil(ACTION_TIMEOUT/1000) then
		print("\127ffffffPlease wait "..math.abs(os.time() - lock - (ACTION_TIMEOUT/1000)).."s before sending another command.")
		return false
	else
		return true
	end
end

-- possibly refresh the lock after the connection is established
-- right now there's a very slight possibility that the lock expires 
-- before the connection drops
-- TODO: the timeout handling should probably go into the player module

-- commands without feedback (play, stop...)
local function Command(player, action)
	if not check_lock() then return end
	lock = os.time()
		
	local timeout_timer = Timer()
	
	local function any_cb(data)
		timeout_timer:Kill()
		player:Disconnect()
		lock = 0
	end
	
	local function ready_cb()
		action()
		player:RequestAny(any_cb)
	end
	
	player:Connect(apcontrol.address, apcontrol.port, apcontrol.timeout, false, ready_cb)

	-- kill the connection if the player doesn't reply before timeout is over
	local function timeout_cb()
		print("timeout")
		player:Disconnect()
	end
	timeout_timer:SetTimeout(ACTION_TIMEOUT, timeout_cb)
end


local function applay_cb()
	local player = LoadPlayer(apcontrol.playername)
	local function action_cb()
		player:Play()
	end
	Command(player, action_cb)
end
RegisterUserCommand("applay", applay_cb)


local function apstop_cb()
	local player = LoadPlayer(apcontrol.playername)
	local function action_cb()
		player:Stop()
	end
	Command(player, action_cb)
end
RegisterUserCommand("apstop", apstop_cb)


local function apnext_cb()
	local player = LoadPlayer(apcontrol.playername)
	local function action_cb()
		player:Next()
	end
	Command(player, action_cb)
end
RegisterUserCommand("apnext", apnext_cb)


local function apprev_cb()
	local player = LoadPlayer(apcontrol.playername)
	local function action_cb()
		player:Prev()
	end
	Command(player, action_cb)
end
RegisterUserCommand("apprev", apprev_cb)


local function aptrack_cb()
	local track = arg and tonumber(arg[1])
	if track then
		player = LoadPlayer(apcontrol.playername)
		local function action_cb()
			player:PlayTrack(track - 1)
		end
		Command(player, action_cb)
	end
end
RegisterUserCommand("aptrack", aptrack_cb)


-- commands with feedback
function title_cb()
	if not check_lock() then return end
	lock = os.time()
	
	local player = LoadPlayer(apcontrol.playername)

	local timeout_timer = Timer()
	
	local function result_cb(title)
		print("Title: "..tostring(title))

		timeout_timer:Kill()
		player:Disconnect()
		lock = 0
	end
	
	local function ready_cb()
		player:RequestTitle(result_cb)
	end

	player:Connect(apcontrol.address, apcontrol.port, apcontrol.timeout, false, ready_cb)

	-- kill the connection if the player doesn't reply before timeout is over
	local function timeout_cb()
		print("timeout")
		player:Disconnect()
	end
	timeout_timer:SetTimeout(ACTION_TIMEOUT, timeout_cb)
end
RegisterUserCommand("aptitle", title_cb)


function playlist_cb()
	if not check_lock() then return end
	lock = os.time()

	local player = LoadPlayer(apcontrol.playername)

	local timeout_timer = Timer()

	local function result_cb(pl)
		for i,v in ipairs(pl) do
			print(i.." "..tostring(v))
		end

		timeout_timer:Kill()	
		player:Disconnect()
		lock = 0
	end
	
	local function ready_cb()
		player:RequestPlaylist(result_cb)
	end

	player:Connect(apcontrol.address, apcontrol.port, apcontrol.timeout, false, ready_cb)
	
	-- kill the connection if the player doesn't reply before timeout is over
	local function timeout_cb()
		print("timeout")
		player:Disconnect()
	end
	timeout_timer:SetTimeout(ACTION_TIMEOUT, timeout_cb)
end
RegisterUserCommand("applaylist", playlist_cb)


function announce_cb()
	if not check_lock() then return end
	lock = os.time()

	local player = LoadPlayer(apcontrol.playername)
	
	local timeout_timer = Timer()

	local function result_cb(title)
		SendChat("/me is listening to "..tostring(title), "CHANNEL")

		timeout_timer:Kill()
		player:Disconnect()
		lock = 0
	end
	
	local function ready_cb()
		player:RequestTitle(result_cb)
	end

	player:Connect(apcontrol.address, apcontrol.port, apcontrol.timeout, false, ready_cb)

	-- kill the connection if the player doesn't reply before timeout is over
	local function timeout_cb()
		print("timeout")
		player:Disconnect()
	end
	timeout_timer:SetTimeout(ACTION_TIMEOUT, timeout_cb)
end
RegisterUserCommand("apannounce", announce_cb)


RegisterUserCommand("apcontrol", apcontrol.Show, apcontrol)

-- sound ducking
local duckinglevel = math.ceil(gksound.GetSoundDuckingLevel() * 100)

local defaultvolume = 100
local muted = false

function apcontrol:OnEvent(name, char, talking)
	if not self.ducking or talking == muted then
		return
	end
	
	local player = LoadPlayer(self.playername)

	if talking then
		if not check_lock() then return end
		lock = os.time()
		
		local timeout_timer = Timer()
	
		local function any_cb(data)
			timeout_timer:Kill()
			player:Disconnect()
			lock = 0
		end
		
		local function result_cb(vol)
			defaultvolume = vol
			muted = true
			player:Volume(duckinglevel)
			player:RequestAny(any_cb)
		end
		
		local function ready_cb()
			player:RequestVolume(result_cb)
		end
		player:Connect(self.address, self.port, self.timeout, false, ready_cb)
		
		-- kill the connection if the player doesn't reply before timeout is over
		local function timeout_cb()
--			print("timeout")
			player:Disconnect()
		end
		timeout_timer:SetTimeout(ACTION_TIMEOUT, timeout_cb)
	else
		local function action_cb()
			player:Volume(defaultvolume)
			muted = false
		end
		Command(player, action_cb)
	end
end

RegisterEvent(apcontrol, "VOICECHAT_PLAYER_TALK_STATUS")


spuck = spuck or {}
spuck.apcontrol = apcontrol
