-- mpd.lua

-- Copyright (c) 2007 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.

-- handler for mpd 

local TCP = dofile("plugins/apcontrol/tcpsock.lua")

local connected = false
local connecttime = 0

local welcome = false
local ready_cb = nil

local _

local mpd = {}

mpd.sock = nil
mpd.contimer = Timer()
mpd.updatetimer = Timer()

-- requestqueue
local requestqueue = {}

local function AddRequest(type, callback)
	table.insert(requestqueue, {type=type, cb=callback})
end

local function RemoveRequest()
	table.remove(requestqueue, 1)
end

local function ResetRequests()
	requestqueue = {}
end

-- actions
function mpd:Stop()
	if welcome then 
		self.sock:Send("stop\n")
	end
end

function mpd:Play()
	if welcome then 
		self.sock:Send("play\n")
	end
end

function mpd:Next()
	if welcome then 
		self.sock:Send("next\n")
	end
end

function mpd:Prev()
	if welcome then 
		self.sock:Send("previous\n")
	end
end

function mpd:PlayTrack(i)
	if welcome and i >= 0 then
		self.sock:Send("play "..i.."\n")
	end
end

function mpd:Volume(i)
	if welcome and i >= 0 and i <= 100 then
		self.sock:Send("setvol "..i.."\n")
	end
end


-- update function
function mpd:Update()
	self:RequestLength(self.Length_cb)
	self:RequestTitle(self.Title_cb)
	self:RequestTime(self.Time_cb)
	self:RequestPlaylist(self.Playlist_cb)
end

-- status callbacks

local function stub()
end

mpd.StatusMsg_cb = stub

mpd.Title_cb = stub
function mpd:RequestTitle(callback)
	if welcome then
		AddRequest("title", callback)
		self.sock:Send("currentsong\n")
	end
end

mpd.Length_cb = stub
function mpd:RequestLength(callback)
	if welcome then 
		AddRequest("length", callback)
		self.sock:Send("status\n")
	end
end

mpd.Time_cb = stub
function mpd:RequestTime(callback)
	if welcome then
		AddRequest("time", callback)
		self.sock:Send("status\n")
	end
end

mpd.Playlist_cb = stub
function mpd:RequestPlaylist(callback)
	if welcome then 
		AddRequest("playlist", callback)
		self.sock:Send("playlistinfo\n")
	end
end

function mpd:RequestVolume(callback)
	if connected then
		AddRequest("volume", callback)
		self.sock:Send("status\n")
	end
end

function mpd:RequestAny(callback)
	if welcome then 
		AddRequest("any", callback)
	end
end

function mpd:ReturnDefaults()
	self.Title_cb("--")
	self.Time_cb(0)
	self.Length_cb(0)
	self.Playlist_cb{}
end

function mpd:LineHandler(line)
	-- don't do anything till we get a welcome line
	if not welcome then
		if string.find(line, "OK MPD .*") == 1 then
			welcome = true
			if ready_cb then ready_cb() end
		else
			self.StatusMsg_cb("Invalid welcome line. Wrong player?")
		end
	elseif #requestqueue > 0 then
		local request = requestqueue[1]

		if request.type == "playlist" then
			request.pl = request.pl or {}

			if string.find(line, "Title: ") == 1 then
				_, _, request.title = string.find(line, "Title: (.*)$")
			end
			
			if string.find(line, "Artist: ") == 1 then
				_, _, request.artist = string.find(line, "Artist: (.*)$")
			end
			
			if string.find(line, "file: ") == 1 then
				_, _, request.file = string.find(line, "file: (.*)$")
			end
				
			if string.find(line, "Id: ") == 1 then
				local titlestr
				if request.title then
					titlestr = (request.artist or "").." - "..request.title
				else
					titlestr = request.file
				end
				table.insert(request.pl, titlestr)
				
				request.title = nil
				request.artist = nil
				request.file = nil
			end
			
			if line == "OK" then
				request.cb(request.pl)
			end
		end

		if request.type == "title" then
			if string.find(line, "Title: ") == 1 then
				_, _, request.title = string.find(line, "Title: (.*)$")
			end
			
			if string.find(line, "Artist: ") == 1 then
				_, _, request.artist = string.find(line, "Artist: (.*)$")
			end

			if string.find(line, "file: ") == 1 then
				_, _, request.file = string.find(line, "file: (.*)$")
			end

			if line == "OK" then
				local titlestr
				if request.title then
					titlestr = (request.artist or "").." - "..request.title
				else
					titlestr = request.file
				end
				request.cb(titlestr)
			end
		end

		if request.type == "time" and string.find(line, "time: ") == 1 then
			string.gsub(line, "time: (.*):.*$", function(match) request.cb(tonumber(match) or 0) end)
		end

		if request.type == "length" and string.find(line, "time: ") == 1 then
			string.gsub(line, "time: .*:(.*)$", function(match) request.cb(tonumber(match) or 0) end)
		end

		if request.type == "volume" and string.find(line, "volume: ") == 1 then
			string.gsub(line, "volume: ([0-9]+)$", function(match) request.cb(tonumber(match) or 0) end)
		end

		if request.type == "any" then
			assert(request.cb)
			request.cb(line)
		end

		if request.type == "any" or line == "OK" or line == "ACK" then 
			RemoveRequest()
		end
	end
end

local function cleanup()
	mpd.updatetimer:Kill()
	connected = false
	welcome = false
	ResetRequests()
end

function mpd:Connect(addr, port, timeout, retry, ready_cb_arg)
	ready_cb = ready_cb_arg

	local function connect_cb(con, err)
		if err then
			cleanup()
			self.StatusMsg_cb(err.." Retrying ..")
			
		elseif con then
			self.StatusMsg_cb("Connected")

			connecttime = os.time()
			connected = true
			self.contimer:Kill()

			if self.Connect_cb then self.Connect_cb() end

			-- go into update loop
			self.updatetimer:SetTimeout(1000, function() self:Update() self.updatetimer:SetTimeout(1000) end)
		end
	end

	local function line_cb(con, line)
		return self:LineHandler(line)
	end

	local function disconnect_cb()
		self.StatusMsg_cb("MPD closed connection")
		cleanup()
		-- restart connect loop
		if retry then self:Connect(addr, port, timeout, retry) end
	end
	
	self:ReturnDefaults()

	if retry then
		if (os.time() - connecttime) > (timeout * 1000) then
			self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
		end

		local function contimer_cb()
			self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
			self.contimer:SetTimeout(timeout)
		end
		self.contimer:SetTimeout(timeout, contimer_cb)
	else
		self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
	end
end

function mpd:Disconnect()
	self.contimer:Kill()

	if self.sock and self.sock.tcp then
		self.sock.tcp:Disconnect()
	end

	cleanup()
	self.StatusMsg_cb("disconnected")
end

return mpd
