-- vlc.lua

-- Copyright (c) 2007 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.

-- handler for vlc's rc module

-- known issues:
-- requests the playlist every second. which may cause a lot of traffic with long lists.
-- it doesn'z check if it's really to a vlc session

local TCP = dofile("plugins/apcontrol/tcpsock.lua")

local connected = false
local connecttime = 0

local vlc = {}

vlc.sock = nil
vlc.contimer = Timer()
vlc.updatetimer = Timer()

-- requestqueue
local requestqueue = {}

local function AddRequest(type, callback)
	table.insert(requestqueue, {type=type, cb=callback})
end

local function RemoveRequest()
	table.remove(requestqueue, 1)
end

local function ResetRequests()
	requestqueue = {}
end

-- actions
function vlc:Stop()
	if connected then 
		self.sock:Send("stop\n")
	end
end

function vlc:Play()
	if connected then 
		self.sock:Send("play\n")
	end
end

function vlc:Next()
	if connected then 
		self.sock:Send("next\n")
	end
end

function vlc:Prev()
	if connected then 
		self.sock:Send("prev\n")
	end
end

function vlc:PlayTrack(i)
	if connected and i >= 0 then 
		self.sock:Send("goto "..(i+1).."\n")
	end
end

function vlc:Volume(i)
	if connected and i >= 0 and i <= 200 then
		local vol = 0
		if i > 0 then
			vol = math.ceil((i / 100) * 256)
		end
		self.sock:Send("volume "..vol.."\n")
	end
end


-- update funtion
function vlc:Update()
	if connected then
		self:RequestTime(self.Time_cb)
		self:RequestPlaylist(self.Playlist_cb)
	end
end

-- status callbacks

local function stub()
end

vlc.StatusMsg_cb = stub

vlc.Title_cb = stub
function vlc:RequestTitle(callback)
	if connected then
		AddRequest("title", callback)
		self.sock:Send("get_title\n")
	end
end

vlc.Length_cb = stub
function vlc:RequestLength(callback)
	if connected then 
		AddRequest("length", callback)
		self.sock:Send("get_length\n")
	end
end

vlc.Time_cb = stub
function vlc:RequestTime(callback)
	if connected then 
		AddRequest("time", callback)
		self.sock:Send("get_time\n")
	end
end

vlc.Playlist_cb = stub
function vlc:RequestPlaylist(callback)
	if connected then 
		AddRequest("playlist", callback)
		self.sock:Send("playlist\n")
	end
end

function vlc:RequestVolume(callback)
	if connected then
		AddRequest("volume", callback)
		self.sock:Send("volume\n")
	end
end

function vlc:RequestAny(callback)
	if connected then 
		AddRequest("any", callback)
	end
end

function vlc:ReturnDefaults()
	if self.Title_cb then self.Title_cb("--") end
	if self.Time_cb then self.Time_cb(0) end
	if self.Length_cb then self.Length_cb(0) end
	if self.Playlist_cb then self.Playlist_cb{} end
end

function vlc:LineHandler(line)
	-- track change
	if string.find(line, "status change: %( new input: ") == 1  then
		self:RequestTitle(self.Title_cb)
		self:RequestLength(self.Length_cb)
	end

	if #requestqueue > 0 then
		local request = requestqueue[1]
		
		if request.type == "any" then
			assert(request.cb)
			request.cb(line)
			RemoveRequest()
			return
		end

		if line:find("status change: %( audio volume: ") == 1 and
		   request.type == "volume"
		then 
			local volume = tonumber(line:match("status change: %( audio volume: ([0-9]+) "))
			if volume then
				if volume > 0 then
					volume = volume / 256 * 100
				end
				request.cb(volume)
			end
			RemoveRequest()
			return
		end

		-- do nothing if this is a status line
		if string.find(line, "status change: %(") == 1 then return end
		
		if request.type == "playlist" then
			request.pl = request.pl or {}
			if string.find(line, "playlist: returned 0") == 1 then
				request.cb(request.pl)
				RemoveRequest()
			else
--				print("playlist: "..tostring(line))
				string.gsub(line, "^|   - (.*)$", function(match) table.insert(request.pl, match) end)
			end
		else
			request.cb(tonumber(line) or line)
			RemoveRequest()
		end
	end
end

local function cleanup()
	vlc.updatetimer:Kill()
	connected = false
	ResetRequests()
end

function vlc:Connect(addr, port, timeout, retry, ready_cb)
	local function connect_cb(con, err)
		if err then
			cleanup()
			self.StatusMsg_cb(err.." Retrying..")
		elseif con then
			self.StatusMsg_cb("Connected")

			connecttime = os.time()
			connected = true
			self.contimer:Kill()

			-- init elements that only update on track change
			self:RequestTitle(self.Title_cb)
			self:RequestLength(self.Length_cb)
			
			if ready_cb then ready_cb() end

			-- go into update loop
			self.updatetimer:SetTimeout(1000, function() self:Update() self.updatetimer:SetTimeout(1000) end)
		end
	end

	local function line_cb(con, line)
		return self:LineHandler(line)
	end

	local function disconnect_cb()
		self.StatusMsg_cb("VLC closed connection")
		cleanup()
		-- restart connect loop
		if retry then self:Connect(addr, port, timeout, retry) end
	end

	self:ReturnDefaults()

	if retry then
		if (os.time() - connecttime) > (timeout * 1000) then
			self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
		end

		local function contimer_cb()
			self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
			self.contimer:SetTimeout(timeout)
		end
		self.contimer:SetTimeout(timeout, contimer_cb)
	else
		self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
	end
end

function vlc:Disconnect()
	self.contimer:Kill()

	if self.sock and self.sock.tcp then
		self.sock.tcp:Disconnect()
	end

	cleanup()
	self.StatusMsg_cb("Disconnected")
end

return vlc
