-- xmms.lua

-- Copyright (c) 2007 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.

-- handler for xmms using inetctl

local TCP = dofile("plugins/apcontrol/tcpsock.lua")

local _

local connected = false
local connecttime = 0
local disconnect = false

local welcome = false
local ready_cb = nil

local xmms = {}

xmms.sock = nil
xmms.contimer = Timer()
xmms.disctimer = Timer()

local loadtime = os.time()

-- requestqueue
local requestqueue = {}

local function AddRequest(type, callback)
	-- only accept requests after we got welcome line
	if welcome then
		table.insert(requestqueue, {type=type, cb=callback})
	end
end

local function RemoveRequest()
	table.remove(requestqueue, 1)
end

local function ResetRequests()
	requestqueue = {}
end


-- actions
function xmms:Stop()
	if welcome then 
		self.sock:Send("STOP\n")
	end
end

function xmms:Play()
	if welcome then 
		self.sock:Send("PLAY\n")
	end
end

function xmms:Next()
	if welcome then 
		self.sock:Send("NEXT_TRACK\n")
	end
end

function xmms:Prev()
	if welcome then 
		self.sock:Send("PREV_TRACK\n")
	end
end

function xmms:PlayTrack(i)
	if welcome and i >= 0 then 
		self.sock:Send("JUMP_TO_TRACK "..i.."\n")
		self:Play()
	end
end

function xmms:Exit()
	if welcome then 
		self.sock:Send("EXIT\n")
	end
end

function xmms:Volume(i)
	if welcome then 
		self.sock:Send("VOLUME "..i.."\n")
	end
end

-- status callbacks

local function stub()
end

xmms.StatusMsg_cb = stub

xmms.Length_cb = stub
function xmms:RequestLength(callback)
	if welcome then 
		self.sock:Send("REQUEST POSITION\n") 
		AddRequest("length", callback)
	end
end

xmms.Time_cb = stub
function xmms:RequestTime(callback)
	if welcome then 
		AddRequest("time", callback)
		self.sock:Send("REQUEST POSITION\n") 
	end
end

xmms.Title_cb = stub
function xmms:RequestTitle(callback)
	if welcome then
		AddRequest("title", callback)
		self.sock:Send("REQUEST TRACK\n") 
	end
end

xmms.Playlist_cb = stub
function xmms:RequestPlaylist(callback)
	if welcome then 
		AddRequest("playlist", callback)
		self.sock:Send("REQUEST PLAYLIST_TRACKS\n")
	end
end

function xmms:RequestVolume(callback)
	if welcome then 
		AddRequest("volume", callback)
	end
end

function xmms:RequestAny(callback)
	if welcome then 
		AddRequest("any", callback)
	end
end

function xmms:ReturnDefaults()
	self.Title_cb("--")
	self.Time_cb(0)
	self.Length_cb(0)
	self.Playlist_cb{}
end

function xmms:LineHandler(line)
	-- XXX: really needed?
	if not connected then
		return
	end
	
	-- don't do anything till we get a welcome line
	if not welcome then
		if string.find(line, "000 .+ XMMS_INETCTL server .*") == 1 then
			welcome = true
			if ready_cb then ready_cb() end
		else
			self.StatusMsg_cb("Invalid welcome line. Wrong player?")
			return
		end
	end

	local request = (#requestqueue > 0) and requestqueue[1] or {}
	local type = string.sub(line, 1, 3)

	if type == "010" then
		local volume
		_, _, volume = string.find(line, "^%d%d%d %S+ ([0-9]*) .*")
		
		if request.type == "volume" then
			request.cb(tonumber(volume))
			RemoveRequest()
		end
	end

	if type == "011" then
		local title
		_, _, title = string.find(line, "^%d%d%d \"(.*)\" \"(.*)\" .*")
		
		self.Title_cb(title)
		if request.type == "title" then
			request.cb(title)
			RemoveRequest()
		end
		self:RequestLength(self.Length_cb)
	end

	if type == "013" then
		local time, length
		_, _, time, length = string.find(line, "^%d%d%d (%d+) (%d+)")

		time = (tonumber(time) or 0) / 1000
		self.Time_cb(time)
		if request.type == "time" then
			request.cb(time)
			RemoveRequest()
		end

		length = (tonumber(length) or 0) / 1000
		if request.type == "length" then
			request.cb(length)
			RemoveRequest()
		end
	end
	
	if type == "014" then
		if request.type == "playlist" then
			request.pl = request.pl or {}
			if string.find(line, "014 END") == 1 then
				request.cb(request.pl)
				RemoveRequest()
			elseif string.find(line, "014 START") ~= 1 then
				string.gsub(line, "^%d%d%d %d+ \"(.*)\" \"(.*)\" .*", function(match) table.insert(request.pl, match) end)
			end
		end
	end

	if type == "012" then
		self:RequestPlaylist(self.Playlist_cb)
	end

	if request.type == "any" then
		assert(request.cb)
		request.cb(line)
		RemoveRequest()
	end
end


local function cleanup()
	connected = false
	welcome = false
	ResetRequests()
end

function xmms:Connect(addr, port, timeout, retry, ready_cb_arg)
	ready_cb = ready_cb_arg

	local function connect_cb(con, err)
		if err then
			cleanup()

			self.sock = nil
			self.disctimer:Kill()

			self.StatusMsg_cb(tostring(err).." Retrying..")
		elseif con then
			self.StatusMsg_cb("Connected")

			connecttime = os.time()
			connected = true
			self.contimer:Kill()
			
			if self.Connect_cb then self.Connect_cb() end

			con:Send("FORMATTED_TIME off\n")
		end
	end

	local function line_cb(con, line)
		return self:LineHandler(line)
	end

	local function disconnect_cb()
		self.StatusMsg_cb("XMMS closed connection")
		cleanup()

		-- disconnect is set when :Disconnect was called
		-- kill the disconnect timer and 
		-- don't try to reconnect
		if disconnect then
			self.disctimer:Kill()
			disconnect = false
			return
		end

		self.sock = nil

		-- restart connect loop
		if retry then
			print("retrying")
			self:Connect(addr, port, timeout, retry)
		end
	end

	self:ReturnDefaults()

	if retry then
		if (os.time() - connecttime) > (timeout * 1000) then
			self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
		end

		local function contimer_cb()
			self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
			self.contimer:SetTimeout(timeout)
		end
		self.contimer:SetTimeout(timeout, contimer_cb)
	else
		self.sock = TCP.make_client(addr, port, connect_cb, line_cb, disconnect_cb)
	end
end

function xmms:Disconnect()
	self.contimer:Kill()
	
	if connected then
		local discsock = self.sock
		local function disctimer_cb()
			discsock.tcp:Disconnect()
		end
		self.disctimer:SetTimeout(2000, disctimer_cb)

		self.sock:Send("EXIT\n")
		disconnect = true
	end

	cleanup()
	self.sock = nil
end

return xmms
